/* pango-script-lang-table.h:
 * 
 * Generated by gen-script-for-lang.c
 * Date: 2015-12-14
 * Source: fontconfig-2.11.94
 * 
 * Do not edit.
 */
typedef struct _PangoScriptForLang {
  const char lang[7];
  PangoScript scripts[3];
} PangoScriptForLang;

static const PangoScriptForLang pango_script_for_lang[] = {
  { "aa",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "ab",     { PANGO_SCRIPT_CYRILLIC/*90*/ } },
  { "af",     { PANGO_SCRIPT_LATIN/*69*/ } },
  { "ak",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "am",     { PANGO_SCRIPT_ETHIOPIC/*264*/ } },
  { "an",     { PANGO_SCRIPT_LATIN/*66*/ } },
  { "ar",     { PANGO_SCRIPT_ARABIC/*36*/ } },
  { "as",     { PANGO_SCRIPT_BENGALI/*64*/ } },
  { "ast",    { PANGO_SCRIPT_LATIN/*70*/ } },
  { "av",     { PANGO_SCRIPT_CYRILLIC/*67*/ } },
  { "ay",     { PANGO_SCRIPT_LATIN/*60*/ } },
  { "az-az",  { PANGO_SCRIPT_LATIN/*66*/ } },
  { "az-ir",  { PANGO_SCRIPT_ARABIC/*38*/ } },
  { "ba",     { PANGO_SCRIPT_CYRILLIC/*82*/ } },
  { "be",     { PANGO_SCRIPT_CYRILLIC/*68*/ } },
  { "ber-dz", { PANGO_SCRIPT_LATIN/*70*/ } },
  { "ber-ma", { PANGO_SCRIPT_TIFINAGH/*32*/ } },
  { "bg",     { PANGO_SCRIPT_CYRILLIC/*60*/ } },
  { "bh",     { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "bho",    { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "bi",     { PANGO_SCRIPT_LATIN/*58*/ } },
  { "bin",    { PANGO_SCRIPT_LATIN/*76*/ } },
  { "bm",     { PANGO_SCRIPT_LATIN/*60*/ } },
  { "bn",     { PANGO_SCRIPT_BENGALI/*63*/ } },
  { "bo",     { PANGO_SCRIPT_TIBETAN/*95*/ } },
  { "br",     { PANGO_SCRIPT_LATIN/*64*/ } },
  { "brx",    { PANGO_SCRIPT_DEVANAGARI/*78*/ } },
  { "bs",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "bua",    { PANGO_SCRIPT_CYRILLIC/*70*/ } },
  { "byn",    { PANGO_SCRIPT_ETHIOPIC/*255*/ } },
  { "ca",     { PANGO_SCRIPT_LATIN/*74*/ } },
  { "ce",     { PANGO_SCRIPT_CYRILLIC/*67*/ } },
  { "ch",     { PANGO_SCRIPT_LATIN/*58*/ } },
  { "chm",    { PANGO_SCRIPT_CYRILLIC/*76*/ } },
  { "chr",    { PANGO_SCRIPT_CHEROKEE/*85*/ } },
  { "co",     { PANGO_SCRIPT_LATIN/*84*/ } },
  { "crh",    { PANGO_SCRIPT_LATIN/*68*/ } },
  { "cs",     { PANGO_SCRIPT_LATIN/*82*/ } },
  { "csb",    { PANGO_SCRIPT_LATIN/*74*/ } },
  { "cu",     { PANGO_SCRIPT_CYRILLIC/*103*/ } },
  { "cv",     { PANGO_SCRIPT_CYRILLIC/*72*/, PANGO_SCRIPT_LATIN/*2*/ } },
  { "cy",     { PANGO_SCRIPT_LATIN/*78*/ } },
  { "da",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "de",     { PANGO_SCRIPT_LATIN/*59*/ } },
  { "doi",    { PANGO_SCRIPT_DEVANAGARI/*81*/ } },
  { "dv",     { PANGO_SCRIPT_THAANA/*49*/ } },
  { "dz",     { PANGO_SCRIPT_TIBETAN/*95*/ } },
  { "ee",     { PANGO_SCRIPT_LATIN/*96*/ } },
  { "el",     { PANGO_SCRIPT_GREEK/*69*/ } },
  { "en",     { PANGO_SCRIPT_LATIN/*72*/ } },
  { "eo",     { PANGO_SCRIPT_LATIN/*64*/ } },
  { "es",     { PANGO_SCRIPT_LATIN/*66*/ } },
  { "et",     { PANGO_SCRIPT_LATIN/*64*/ } },
  { "eu",     { PANGO_SCRIPT_LATIN/*56*/ } },
  { "fa",     { PANGO_SCRIPT_ARABIC/*38*/ } },
  { "fat",    { PANGO_SCRIPT_LATIN/*70*/ } },
  { "ff",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "fi",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "fil",    { PANGO_SCRIPT_LATIN/*84*/ } },
  { "fj",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "fo",     { PANGO_SCRIPT_LATIN/*68*/ } },
  { "fr",     { PANGO_SCRIPT_LATIN/*84*/ } },
  { "fur",    { PANGO_SCRIPT_LATIN/*66*/ } },
  { "fy",     { PANGO_SCRIPT_LATIN/*75*/ } },
  { "ga",     { PANGO_SCRIPT_LATIN/*80*/ } },
  { "gd",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "gez",    { PANGO_SCRIPT_ETHIOPIC/*218*/ } },
  { "gl",     { PANGO_SCRIPT_LATIN/*66*/ } },
  { "gn",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "gu",     { PANGO_SCRIPT_GUJARATI/*68*/ } },
  { "gv",     { PANGO_SCRIPT_LATIN/*54*/ } },
  { "ha",     { PANGO_SCRIPT_LATIN/*60*/ } },
  { "haw",    { PANGO_SCRIPT_LATIN/*62*/ } },
  { "he",     { PANGO_SCRIPT_HEBREW/*27*/ } },
  { "hi",     { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "hne",    { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "ho",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "hr",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "hsb",    { PANGO_SCRIPT_LATIN/*72*/ } },
  { "ht",     { PANGO_SCRIPT_LATIN/*56*/ } },
  { "hu",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "hy",     { PANGO_SCRIPT_ARMENIAN/*77*/ } },
  { "hz",     { PANGO_SCRIPT_LATIN/*56*/ } },
  { "ia",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "id",     { PANGO_SCRIPT_LATIN/*54*/ } },
  { "ie",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "ig",     { PANGO_SCRIPT_LATIN/*58*/ } },
  { "ii",     { PANGO_SCRIPT_YI/*1165*/ } },
  { "ik",     { PANGO_SCRIPT_CYRILLIC/*68*/ } },
  { "io",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "is",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "it",     { PANGO_SCRIPT_LATIN/*72*/ } },
  { "iu",     { PANGO_SCRIPT_CANADIAN_ABORIGINAL/*161*/ } },
  { "ja",     { PANGO_SCRIPT_HAN/*2134*/, PANGO_SCRIPT_KATAKANA/*88*/, PANGO_SCRIPT_HIRAGANA/*85*/ } },
  { "jv",     { PANGO_SCRIPT_LATIN/*56*/ } },
  { "ka",     { PANGO_SCRIPT_GEORGIAN/*33*/ } },
  { "kaa",    { PANGO_SCRIPT_CYRILLIC/*78*/ } },
  { "kab",    { PANGO_SCRIPT_LATIN/*70*/ } },
  { "ki",     { PANGO_SCRIPT_LATIN/*56*/ } },
  { "kj",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "kk",     { PANGO_SCRIPT_CYRILLIC/*77*/ } },
  { "kl",     { PANGO_SCRIPT_LATIN/*81*/ } },
  { "km",     { PANGO_SCRIPT_KHMER/*63*/ } },
  { "kn",     { PANGO_SCRIPT_KANNADA/*70*/ } },
  { "ko",     { PANGO_SCRIPT_HANGUL/*2442*/ } },
  { "kok",    { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "kr",     { PANGO_SCRIPT_LATIN/*56*/ } },
  { "ks",     { PANGO_SCRIPT_ARABIC/*33*/ } },
  { "ku-am",  { PANGO_SCRIPT_CYRILLIC/*64*/ } },
  { "ku-iq",  { PANGO_SCRIPT_ARABIC/*32*/ } },
  { "ku-ir",  { PANGO_SCRIPT_ARABIC/*32*/ } },
  { "ku-tr",  { PANGO_SCRIPT_LATIN/*62*/ } },
  { "kum",    { PANGO_SCRIPT_CYRILLIC/*66*/ } },
  { "kv",     { PANGO_SCRIPT_CYRILLIC/*70*/ } },
  { "kw",     { PANGO_SCRIPT_LATIN/*64*/ } },
  { "kwm",    { PANGO_SCRIPT_LATIN/*52*/ } },
  { "ky",     { PANGO_SCRIPT_CYRILLIC/*70*/ } },
  { "la",     { PANGO_SCRIPT_LATIN/*68*/ } },
  { "lah",    { PANGO_SCRIPT_ARABIC/*27*/ } },
  { "lb",     { PANGO_SCRIPT_LATIN/*75*/ } },
  { "lez",    { PANGO_SCRIPT_CYRILLIC/*67*/ } },
  { "lg",     { PANGO_SCRIPT_LATIN/*54*/ } },
  { "li",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "ln",     { PANGO_SCRIPT_LATIN/*78*/ } },
  { "lo",     { PANGO_SCRIPT_LAO/*55*/ } },
  { "lt",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "lv",     { PANGO_SCRIPT_LATIN/*78*/ } },
  { "mai",    { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "mg",     { PANGO_SCRIPT_LATIN/*56*/ } },
  { "mh",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "mi",     { PANGO_SCRIPT_LATIN/*64*/ } },
  { "mk",     { PANGO_SCRIPT_CYRILLIC/*42*/ } },
  { "ml",     { PANGO_SCRIPT_MALAYALAM/*68*/ } },
  { "mn-cn",  { PANGO_SCRIPT_MONGOLIAN/*130*/ } },
  { "mn-mn",  { PANGO_SCRIPT_CYRILLIC/*70*/ } },
  { "mni",    { PANGO_SCRIPT_BENGALI/*75*/ } },
  { "mo",     { PANGO_SCRIPT_CYRILLIC/*66*/, PANGO_SCRIPT_LATIN/*62*/ } },
  { "mr",     { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "ms",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "mt",     { PANGO_SCRIPT_LATIN/*72*/ } },
  { "my",     { PANGO_SCRIPT_MYANMAR/*48*/ } },
  { "na",     { PANGO_SCRIPT_LATIN/*60*/ } },
  { "nb",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "nds",    { PANGO_SCRIPT_LATIN/*59*/ } },
  { "ne",     { PANGO_SCRIPT_DEVANAGARI/*70*/ } },
  { "ng",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "nl",     { PANGO_SCRIPT_LATIN/*82*/ } },
  { "nn",     { PANGO_SCRIPT_LATIN/*76*/ } },
  { "no",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "nqo",    { PANGO_SCRIPT_NKO/*59*/ } },
  { "nr",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "nso",    { PANGO_SCRIPT_LATIN/*58*/ } },
  { "nv",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "ny",     { PANGO_SCRIPT_LATIN/*54*/ } },
  { "oc",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "om",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "or",     { PANGO_SCRIPT_ORIYA/*68*/ } },
  { "os",     { PANGO_SCRIPT_CYRILLIC/*66*/ } },
  { "ota",    { PANGO_SCRIPT_ARABIC/*37*/ } },
  { "pa",     { PANGO_SCRIPT_GURMUKHI/*63*/ } },
  { "pa-pk",  { PANGO_SCRIPT_ARABIC/*27*/ } },
  { "pap-an", { PANGO_SCRIPT_LATIN/*72*/ } },
  { "pap-aw", { PANGO_SCRIPT_LATIN/*54*/ } },
  { "pl",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "ps-af",  { PANGO_SCRIPT_ARABIC/*49*/ } },
  { "ps-pk",  { PANGO_SCRIPT_ARABIC/*49*/ } },
  { "pt",     { PANGO_SCRIPT_LATIN/*82*/ } },
  { "qu",     { PANGO_SCRIPT_LATIN/*54*/ } },
  { "quz",    { PANGO_SCRIPT_LATIN/*54*/ } },
  { "rm",     { PANGO_SCRIPT_LATIN/*66*/ } },
  { "rn",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "ro",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "ru",     { PANGO_SCRIPT_CYRILLIC/*66*/ } },
  { "rw",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "sa",     { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "sah",    { PANGO_SCRIPT_CYRILLIC/*76*/ } },
  { "sat",    { PANGO_SCRIPT_DEVANAGARI/*68*/ } },
  { "sc",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "sco",    { PANGO_SCRIPT_LATIN/*56*/ } },
  { "sd",     { PANGO_SCRIPT_ARABIC/*54*/ } },
  { "se",     { PANGO_SCRIPT_LATIN/*66*/ } },
  { "sel",    { PANGO_SCRIPT_CYRILLIC/*66*/ } },
  { "sg",     { PANGO_SCRIPT_LATIN/*72*/ } },
  { "sh",     { PANGO_SCRIPT_CYRILLIC/*94*/, PANGO_SCRIPT_LATIN/*62*/ } },
  { "shs",    { PANGO_SCRIPT_LATIN/*46*/ } },
  { "si",     { PANGO_SCRIPT_SINHALA/*73*/ } },
  { "sid",    { PANGO_SCRIPT_ETHIOPIC/*281*/ } },
  { "sk",     { PANGO_SCRIPT_LATIN/*86*/ } },
  { "sl",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "sm",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "sma",    { PANGO_SCRIPT_LATIN/*60*/ } },
  { "smj",    { PANGO_SCRIPT_LATIN/*60*/ } },
  { "smn",    { PANGO_SCRIPT_LATIN/*68*/ } },
  { "sms",    { PANGO_SCRIPT_LATIN/*80*/ } },
  { "sn",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "so",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "sq",     { PANGO_SCRIPT_LATIN/*56*/ } },
  { "sr",     { PANGO_SCRIPT_CYRILLIC/*60*/ } },
  { "ss",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "st",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "su",     { PANGO_SCRIPT_LATIN/*54*/ } },
  { "sv",     { PANGO_SCRIPT_LATIN/*68*/ } },
  { "sw",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "syr",    { PANGO_SCRIPT_SYRIAC/*45*/ } },
  { "ta",     { PANGO_SCRIPT_TAMIL/*48*/ } },
  { "te",     { PANGO_SCRIPT_TELUGU/*70*/ } },
  { "tg",     { PANGO_SCRIPT_CYRILLIC/*78*/ } },
  { "th",     { PANGO_SCRIPT_THAI/*73*/ } },
  { "ti-er",  { PANGO_SCRIPT_ETHIOPIC/*255*/ } },
  { "ti-et",  { PANGO_SCRIPT_ETHIOPIC/*281*/ } },
  { "tig",    { PANGO_SCRIPT_ETHIOPIC/*221*/ } },
  { "tk",     { PANGO_SCRIPT_LATIN/*68*/ } },
  { "tl",     { PANGO_SCRIPT_LATIN/*84*/ } },
  { "tn",     { PANGO_SCRIPT_LATIN/*58*/ } },
  { "to",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "tr",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "ts",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "tt",     { PANGO_SCRIPT_CYRILLIC/*76*/ } },
  { "tw",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "ty",     { PANGO_SCRIPT_LATIN/*64*/ } },
  { "tyv",    { PANGO_SCRIPT_CYRILLIC/*70*/ } },
  { "ug",     { PANGO_SCRIPT_ARABIC/*33*/ } },
  { "uk",     { PANGO_SCRIPT_CYRILLIC/*72*/ } },
  { "ur",     { PANGO_SCRIPT_ARABIC/*27*/ } },
  { "uz",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "ve",     { PANGO_SCRIPT_LATIN/*62*/ } },
  { "vi",     { PANGO_SCRIPT_LATIN/*186*/ } },
  { "vo",     { PANGO_SCRIPT_LATIN/*54*/ } },
  { "vot",    { PANGO_SCRIPT_LATIN/*62*/ } },
  { "wa",     { PANGO_SCRIPT_LATIN/*70*/ } },
  { "wal",    { PANGO_SCRIPT_ETHIOPIC/*281*/ } },
  { "wen",    { PANGO_SCRIPT_LATIN/*76*/ } },
  { "wo",     { PANGO_SCRIPT_LATIN/*66*/ } },
  { "xh",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "yap",    { PANGO_SCRIPT_LATIN/*58*/ } },
  { "yi",     { PANGO_SCRIPT_HEBREW/*27*/ } },
  { "yo",     { PANGO_SCRIPT_LATIN/*114*/ } },
  { "za",     { PANGO_SCRIPT_LATIN/*52*/ } },
  { "zh-cn",  { PANGO_SCRIPT_HAN/*6763*/ } },
  { "zh-hk",  { PANGO_SCRIPT_HAN/*1083*/ } },
  { "zh-mo",  { PANGO_SCRIPT_HAN/*1083*/ } },
  { "zh-sg",  { PANGO_SCRIPT_HAN/*6763*/ } },
  { "zh-tw",  { PANGO_SCRIPT_HAN/*13063*/ } },
  { "zu",     { PANGO_SCRIPT_LATIN/*52*/ } }
};
